from _random import Random
import difflib
from inspect import isgenerator
import os
import pprint
import unittest
from unittest.util import safe_repr, _common_shorten_repr

from hwt.hdlObjects.constants import Time
from hwt.hdlObjects.types.arrayVal import ArrayVal
from hwt.hdlObjects.value import Value
from hwt.simulator.agentConnector import valToInt
from hwt.simulator.configVhdlTestbench import HdlSimConfigVhdlTestbench
from hwt.simulator.hdlSimulator import HdlSimulator
from hwt.simulator.shortcuts import simPrepare
from hwt.simulator.simSignal import SimSignal
from hwt.simulator.utils import agent_randomize
from hwt.simulator.vcdHdlSimConfig import VcdHdlSimConfig


def allValuesToInts(sequenceOrVal):
    if isinstance(sequenceOrVal, ArrayVal):
        sequenceOrVal = sequenceOrVal.val

    if isinstance(sequenceOrVal, Value):
        return valToInt(sequenceOrVal)
    elif not sequenceOrVal:
        return sequenceOrVal
    elif isinstance(sequenceOrVal, (list, tuple)) or isgenerator(sequenceOrVal):
        l = []
        for i in sequenceOrVal:
            l.append(allValuesToInts(i))

        if isinstance(sequenceOrVal, tuple):
            return tuple(l)

        return l
    else:
        return sequenceOrVal


class SimTestCase(unittest.TestCase):
    """
    This is TestCase class contains methods which are usually used during
    hdl simulation.

    :attention: self.model, self.procs has to be specified before running doSim (you can use prepareUnit method)
    """
    _defaultSeed = 317
    _rand = Random(_defaultSeed)

    def getTestName(self):
        className, testName = self.id().split(".")[-2:]
        return "%s_%s" % (className, testName)

    def doSim(self, time):
        outputFileName = "tmp/" + self.getTestName() + ".vcd"
        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            # return _simUnitVcd(simModel, stimulFunctions, outputFile=f, time=time)
            sim = HdlSimulator()

            # configure simulator to log in vcd
            sim.config = VcdHdlSimConfig(outputFile)

            # run simulation, stimul processes are register after initial initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs)
            return sim

    def dumpHdlTestbench(self, time, file=None):
        if file:
            outputFileName = file
        else:
            outputFileName = "tmp/" + self.getTestName() + "_tb.vhd"
        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            # return _simUnitVcd(simModel, stimulFunctions, outputFile=f, time=time)
            sim = HdlSimulator()

            # configure simulator to log in vcd
            sim.config = HdlSimConfigVhdlTestbench(self.u)

            # run simulation, stimul processes are register after initial initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs)

            sim.config.dump(outputFile)

            return sim

    def assertValEqual(self, first, second, msg=None):
        if isinstance(first, SimSignal):
            first = first._val
        if not isinstance(first, int) and first is not None:
            first = valToInt(first)

        return unittest.TestCase.assertEqual(self, first, second, msg=msg)

    def assertEmpty(self, val, msg=None):
        return unittest.TestCase.assertEqual(self, len(val), 0, msg=msg)

    def assertValSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """
        An equality assertion for ordered sequences (like lists and tuples).
        For the purposes of this function, a valid ordered sequence type is one
        which can be indexed, has a length, and has an equality operator.

        Args:

        :param seq1: can contain instance of values or nested list of them
        :param seq2: items are not converted
        :param seq_type: The expected datatype of the sequences, or None if no
            datatype should be enforced.
        :param msg: Optional message to use on failure instead of a list of
            differences.
        """
        seq1 = allValuesToInts(seq1)
        if seq_type is not None:
            seq_type_name = seq_type.__name__
            if not isinstance(seq1, seq_type):
                raise self.failureException('First sequence is not a %s: %s'
                                            % (seq_type_name, safe_repr(seq1)))
            if not isinstance(seq2, seq_type):
                raise self.failureException('Second sequence is not a %s: %s'
                                            % (seq_type_name, safe_repr(seq2)))
        else:
            seq_type_name = "sequence"

        differing = None
        try:
            len1 = len(seq1)
        except (TypeError, NotImplementedError):
            differing = 'First %s has no length.    Non-sequence?' % (
                    seq_type_name)

        if differing is None:
            try:
                len2 = len(seq2)
            except (TypeError, NotImplementedError):
                differing = 'Second %s has no length.    Non-sequence?' % (
                        seq_type_name)

        if differing is None:
            if seq1 == seq2:
                return

            differing = '%ss differ: %s != %s\n' % (
                    (seq_type_name.capitalize(),) +
                    _common_shorten_repr(seq1, seq2))

            for i in range(min(len1, len2)):
                try:
                    item1 = seq1[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of first %s\n'
                                  % (i, seq_type_name))
                    break

                try:
                    item2 = seq2[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of second %s\n'
                                  % (i, seq_type_name))
                    break

                if item1 != item2:
                    differing += ('\nFirst differing element %d:\n%s\n%s\n'
                                  % ((i,) + _common_shorten_repr(item1, item2)))
                    break
            else:
                if (len1 == len2 and seq_type is None and
                   type(seq1) != type(seq2)):
                    # The sequences are the same, but have differing types.
                    return

            if len1 > len2:
                differing += ('\nFirst %s contains %d additional '
                              'elements.\n' % (seq_type_name, len1 - len2))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len2, safe_repr(seq1[len2])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of first %s\n' % (len2, seq_type_name))
            elif len1 < len2:
                differing += ('\nSecond %s contains %d additional '
                              'elements.\n' % (seq_type_name, len2 - len1))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len1, safe_repr(seq2[len1])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of second %s\n' % (len1, seq_type_name))
        standardMsg = differing
        diffMsg = '\n' + '\n'.join(
            difflib.ndiff(pprint.pformat(seq1).splitlines(),
                          pprint.pformat(seq2).splitlines()))

        standardMsg = self._truncateMessage(standardMsg, diffMsg)
        msg = self._formatMessage(msg, standardMsg)
        self.fail(msg)

    def randomize(self, intf):
        """
        Randomly disable and enable interface for testing purposes
        """
        self.procs.append(agent_randomize(intf._ag,
                                          50 * Time.ns,
                                          seed=self._rand.getrandbits(64)))

    def prepareUnit(self, unit, modelCls=None, dumpModelIn=None, onAfterToRtl=None):
        """
        Create simulation model and connect it with interfaces of original unit
        and decorate it with agents and collect all simulation processes

        :param unit: interface level unit which you wont prepare for simulation
        :param modelCls: class of rtl simulation model to run simulation on, if is None
            rtl sim model will be generated from unit
        :param dumpModelIn: folder to where put sim model files (if is None sim model will be constructed only in memory)
        :param onAfterToRtl: callback fn(unit) which will be called unit after it will
            be synthesised to rtl
        """
        self.u, self.model, self.procs = simPrepare(unit,
                                                    modelCls=modelCls,
                                                    dumpModelIn=dumpModelIn,
                                                    onAfterToRtl=onAfterToRtl)

    def setUp(self):
        self._rand.seed(self._defaultSeed)



from hwt.hdlObjects.types.hdlType import HdlType


class HStream(HdlType):
    """
    Stream data type
    stream is like array with unlimited size
    """
    def __init__(self, elmType):
        """
        :param elmType: type of item used for alignment purposes
        :attention: width of itemType and width of interface over which
            is this stream send are two different things
        """
        self.elmType = elmType

    def __hash__(self):
        return hash(self.elmType)



from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.statements import IfContainer, SwitchContainer
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.value import Value
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.termUsageResolver import getBaseCond
from hwt.hdlObjects.types.bits import Bits


# (max count of elsifs with eq on same variable)
SWITCH_THRESHOLD = 2


def condWithoutResolved(cond, resolvedCnt):
    return reversed(cond[:len(cond) - resolvedCnt])


def splitStatementsOnCond(statements, resolvedCondCnt):
    # resolve how many condition items can we take into into actual if statement
    simplestStm = statements[0]
    cntOfSameConditions = len(simplestStm.cond)
    for a in statements:
        _cntOfSameConditions = 0
        for c0, c1 in zip(condWithoutResolved(a.cond, resolvedCondCnt),
                          condWithoutResolved(simplestStm.cond, resolvedCondCnt)):
            if c0 is c1:
                _cntOfSameConditions += 1
            else:
                break

        if _cntOfSameConditions < cntOfSameConditions:
            cntOfSameConditions = _cntOfSameConditions

        if len(a.cond) < len(simplestStm.cond):
            simplestStm = a

    assert cntOfSameConditions >= 0
    ifFalse = []
    independent = []
    if cntOfSameConditions > 0:
        # if cond:
        #   stm
        l = len(simplestStm.cond)
        topCond = simplestStm.cond[l - resolvedCondCnt - cntOfSameConditions:l - resolvedCondCnt]
        ifTrue = list(renderIfTree(statements, resolvedCondCnt + cntOfSameConditions))
    else:
        # if cond:
        #    stm
        # else:
        #    stm
        condIndx = -resolvedCondCnt - 1
        topCond, _ = getBaseCond(simplestStm.cond[condIndx])
        topCondNeg = ~topCond

        ifTrue = []
        for s in statements:
            c = s.cond[condIndx]
            if c is topCond:
                ifTrue.append(s)
            elif c is topCondNeg:
                ifFalse.append(s)
            else:
                independent.append(s)
        #raise NotImplementedError("Can not use statement ", s,
        #                          " because its position in if-tree is not decisable",
        #                          " topCond:", topCond, " c:", c)
        ifTrue = list(renderIfTree(ifTrue, resolvedCondCnt + 1))
        ifFalse = list(renderIfTree(ifFalse, resolvedCondCnt + 1))
        topCond = [topCond, ]

    assert topCond
    return topCond, ifTrue, ifFalse, independent


def disolveConditionAsEq(condList):
    """
    detect if condition is in format signal == value
    :return: None if condList is not in format [ a == val] or [val == a]
        else return tuple (a, val)
    """
    try:
        if len(condList) > 1:
            raise MultipleDriversExc()

        op = condList[0].singleDriver()
        if isinstance(op, Operator) and op.operator == AllOps.EQ:
            op0 = op.ops[0]
            op1 = op.ops[1]
            if isinstance(op0, RtlSignalBase) and isinstance(op1, Value):
                return op0, op1
            elif isinstance(op1, RtlSignalBase) and isinstance(op0, Value):
                return op1, op0
    except MultipleDriversExc:
        pass


def typeDomainSize(t):
    if isinstance(t, Enum):
        return len(t._allValues)
    elif isinstance(t, Bits):
        return 2 ** t.bit_length()
    else:
        raise TypeError(t)


def renderIfTree(statements, resolvedCnt=0):
    """
    Walk assignments and resolve if tree from conditions
    """

    # filter statements which are not under any condition
    _statements = []
    for a in statements:
        if len(a.cond) > resolvedCnt:
            _statements.append(a)
        else:
            yield a

    if _statements:
        topCond, ifTrue, ifFalse, independent = splitStatementsOnCond(statements, resolvedCnt)
        if independent:
            yield from renderIfTree(independent, resolvedCnt)

        # reduce if else to elifs or switch
        elIfs = []
        while len(ifFalse) == 1:
            stm = ifFalse[0]
            if isinstance(stm, IfContainer):
                elIfs.extend(stm.elIfs)
                elIfs.insert(0, (stm.cond, stm.ifTrue))
                ifFalse = stm.ifFalse
                continue
            elif isinstance(stm, SwitchContainer):
                # try to extend switch statement
                dis = disolveConditionAsEq(topCond)
                if dis:
                    switchOn, topVal = dis
                    if stm.switchOn is switchOn:
                        assert not elIfs
                        stm.cases.insert(0, (topVal, ifTrue))

                        setDefault = False
                        try:
                            setDefault = not stm.default and typeDomainSize(switchOn._dtype) == len(stm.cases)
                        except TypeError:
                            pass
                        if setDefault:
                            # we can use any item as default because switch contains case for
                            # every possible value from type
                            c = stm.cases.pop()
                            stm.default = c[1]
                        yield stm
                        return
            break

        if len(elIfs) >= SWITCH_THRESHOLD:
            # try to convert elifs to switch
            cases = []
            disolvedTopCond = disolveConditionAsEq(topCond)
            if disolvedTopCond is not None:
                switchOn, topVal = disolvedTopCond
                cases.append((topVal, ifTrue))
                canBeConvertedToSwitch = True
                for elIf in elIfs:
                    dis = disolveConditionAsEq(elIf[0])
                    if dis is None or dis[0] is not switchOn:
                        canBeConvertedToSwitch = False
                        break
                    else:
                        cases.append((dis[1], elIf[1]))

                # if only last can not be part of the switch case it can be default
                if not canBeConvertedToSwitch and len(elIfs) == len(cases):
                    default = elIfs[-1]
                    ifFalse = [IfContainer(default[0], default[1], ifFalse)]
                    canBeConvertedToSwitch = True

                if canBeConvertedToSwitch:
                    t = switchOn._dtype
                    # if nothing else and we have enum ad we used all the values
                    if not ifFalse and isinstance(t, Enum) and len(t._allValues) == len(cases):
                        # convert last to default, because hdl languages usually need this
                        ifFalse = cases[-1][1]
                        cases = cases[:-1]
                    default = ifFalse

                    yield SwitchContainer(switchOn, cases, default)
                    return

        yield IfContainer(topCond,
                          ifTrue,
                          ifFalse,
                          elIfs)



from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.value import Value


class HStructField(object):
    def __init__(self, typ, name, info=None):
        assert isinstance(name, str) or name is None, name
        assert isinstance(typ, HdlType),typ
        self.name = name
        self.type = typ
        self.info = info

    def __hash__(self):
        return hash((self.type, self.name))


class HStruct(HdlType):
    """
    C-like structure type
    """
    def __init__(self, *template, name=None):
        """
        :param template: list of tuples (type, name) or HStructField objects
            name can be None it means that there is space in structure
        :param name: optional name used for debugging purposes
        """
        self.fields = []
        self.name = name
        fieldNames = []
        self.isFixedSize = True

        for f in template:
            try:
                field = HStructField(*f)
            except TypeError:
                field = f
            if not isinstance(field, HStructField):
                raise TypeError("Template for struct field %s is not in valid format" % repr(f))
            
            
            self.fields.append(field)
            if field.name is not None:
                fieldNames.append(field.name)

            t = field.type
            if self.isFixedSize:
                if (isinstance(t, HStruct) and not t.isFixedSize)\
                              or not hasattr(t, "bit_length"):
                    self.isFixedSize = False

        self.fields = tuple(self.fields)

        class StructVal(Value):
            __structType = self
            __slots__ = fieldNames

            def __init__(self):
                for f in self.__structType.fields:
                    if f.name is not None:
                        setattr(self, f.name, None)

            def __str__(self):
                buff = ["{"]
                for f in self.__structType.fields:
                    if f.name is not None:
                        val = getattr(self, f.name)
                        buff.append("    %s %r" % (f.name, val))
                buff.append("}")
                return "\n".join(buff)

        self.valueCls = StructVal

    def bit_length(self):
        if self.isFixedSize:
            return sum(map(lambda f: f.type.bit_length(), self.fields))
        else:
            raise TypeError("Can not request bit_lenght on size which has not fixed size")
    
    def getValueCls(self):
        return self.valueCls

    def sizeof(self):
        """
        get size of struct in bytes
        """
        s = 0
        for f in self.template:
            s += f.type.bit_length()

        if s % 8 == 0:
            return s // 8
        else:
            return s // 8 + 1
    
    def __hash__(self):
        return hash((self.name, self.fields))

    def __add__(self, other):
        """
        override of addition, merge struct into one
        """
        assert isinstance(other, HStruct)
        return HStruct(*self.fields, *other.fields)

    def __str__(self):
        if self.name:
            name = self.name + " "
        else:
            name = ""

        buff = ["struct %s{" % name]
        for f in self.fields:
            if f.name is None:
                buff.append("    //%r empty space" % (f.type))
            else:
                buff.append("    %r %s" % (f.type, f.name))

        buff.append("}")
        return "\n".join(buff)



from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.struct import HStructField, HStruct


class FrameTemplateItem(object):
    def __init__(self, name, typ, inFrameBitOffset, internalInterface, externalInterface):
        self.name = name
        self.type = typ

        self.inFrameBitOffset = inFrameBitOffset
        # list of tuples (word index, inWordBitUpper, inWordBitLower)
        self.appearsInWords = []

        self.internalInterface = internalInterface
        self.externalInterface = externalInterface


class FrameTemplate(list):

    def walkWords(self):
        """
        Walks frame over words

        :return: generator of tuples (wordIndex, list of FrameTemplateItems)
        """
        wordRecord = []
        actualWord = 0
        if not self[0].appearsInWords:
                raise Exception("resolveFieldPossitionsInFrame has to be called first")
        for fi in self:
            for w, _, _ in fi.appearsInWords:
                if w == actualWord:
                    wordRecord.append(fi)
                elif w > actualWord:
                    yield actualWord, wordRecord
                    wordRecord = [fi, ]
                    actualWord = w
                else:
                    raise NotImplementedError("Input frame info has to be sorted")
    
        if wordRecord:
            yield actualWord, wordRecord
    
    @classmethod
    def fromHStruct(cls, structT):
        self = cls()
        inFrameOffset = 0
        for f in structT.fields:
            fi = FrameTemplateItem(f.name, f.type, inFrameOffset, None, None)
            self.append(fi)
            inFrameOffset += f.type.bit_length()
        return self    
    
    def resolveFieldPossitionsInFrame(self, dataWidth, disolveArrays=True):
        """
        Format fields in datawords

        :return: number of words
        """
        assert isinstance(dataWidth, int), dataWidth
        bitAddr = 0
        for item in self:
            lower = item.inFrameBitOffset % dataWidth
            remInFirstWord = dataWidth - (item.inFrameBitOffset % dataWidth)
            w = item.type.bit_length()
            
            if isinstance(item.type, Array) and not disolveArrays:
                assert lower == 0, "Start of array has to be aligned"
                item.appearsInWords.append((bitAddr // dataWidth, w, 0))
                bitAddr += w
            else:
                # align start
                if w <= remInFirstWord:
                    upper = lower + w
    
                item.appearsInWords.append((bitAddr // dataWidth, upper, lower))
                if w <= remInFirstWord:
                    bitAddr += w
                    continue
    
                if w // dataWidth > 0:
                    # take aligned middle of field
                    for _ in range(w // dataWidth):
                        item.appearsInWords.append(
                                                    (bitAddr // dataWidth, dataWidth, 0)
                                                    )
                        bitAddr += dataWidth
    
                if w != 0:
                    # take reminder at the end
                    item.appearsInWords.append(
                                                (bitAddr // dataWidth, w, 0)
                                                )
                    bitAddr += w
        i = bitAddr // dataWidth
        if bitAddr % dataWidth == 0:
            return i
        else:
            return i + 1


class StructFieldPart():
    def __init__(self, wordIndex, busWordBitRange, inFieldBitRange):
        """ranges in little endian notation"""
        self.wordIndex = wordIndex
        self.busWordBitRange = busWordBitRange
        self.inFieldBitRange = inFieldBitRange

    def getSignal(self, busDataSignal):
        if self.busWordBitRange == (busDataSignal._dtype.bit_length(), 0):
            return busDataSignal
        else:
            return busDataSignal[self.busWordBitRange[0]:self.busWordBitRange[1]]

    def __repr__(self):
        return "<StructFieldPart, wordIndex:%d, busWordBitRange:%r, inFieldBitRange:%r>" % (
                self.wordIndex, self.busWordBitRange, self.inFieldBitRange)


class StructFieldInfo():
    def __init__(self, typ, name):
        self.type = typ
        self.name = name
        self.interface = None
        self.parts = []

    def discoverFieldParts(self, busDataWidth, startBitIndex):
        """Some fields has to be internally split due data-width of bus,
        there we discover how to split field to words on bus

        :param startWordIndex: bit index where field starts, (f.e. 16 for f2 in struct {uint16_t f1, uint16_t f2})

        """
        fieldWidth = self.type.bit_length()
        inFieldOffset = 0

        while fieldWidth != 0:
            wordIndex = startBitIndex // busDataWidth
            bitRangeOfWord = (busDataWidth * (wordIndex + 1), startBitIndex)  # little-endian
            widthOfPart = min(bitRangeOfWord[0], startBitIndex + fieldWidth) - startBitIndex
            busWordBitOffset = startBitIndex % busDataWidth

            busWordBitRange = (busWordBitOffset + widthOfPart, busWordBitOffset)
            inFieldRange = (inFieldOffset + widthOfPart, inFieldOffset)

            fp = StructFieldPart(wordIndex, busWordBitRange, inFieldRange)
            self.parts.append(fp)

            inFieldOffset += widthOfPart
            startBitIndex += widthOfPart
            fieldWidth -= widthOfPart

        return startBitIndex + fieldWidth

    def __repr__(self):
        return "<StructFieldInfo %s, %r, parts:%s  >" % (
            self.name, self.type, "\n".join(["%r" % s for s in self.parts]))


class StructBusBurstInfo():
    """
    Represents data chunks(bursts) which should be send over interface
    container of parts of fields form struct
    """
    def __init__(self, addrOffset, fieldInfos):
        """
        :param addrOffset: offset of this burst in number of words
        :param fieldInfos: iterable of field StructFieldInfo
        """
        self.addrOffset = addrOffset
        self.fieldInfos = fieldInfos

    def wordCnt(self):
        firstWordIndex = self.fieldInfos[0].parts[0].wordIndex
        lastWordIndex = self.fieldInfos[-1].parts[-1].wordIndex

        return lastWordIndex - firstWordIndex + 1

    @staticmethod
    def packFieldInfosToBusBurst(structInfos, maxDummyWords, wordIndexToAddrRatio):
        """
        :param structInfos: iterable of StructFieldInfo which are describing target structure
        :param maxDummyWords: maximum allowable not used words in bus burst
        :param wordIndexToAddrRatio: addrOffset = wordIndex of field * wordIndexToAddrRatio
        :return: list of StructBusBurstInfo
        """
        busBursts = []

        try:
            firstWordIndex = structInfos[0].parts[0].wordIndex
        except IndexError:
            return []

        lastWordIndex = firstWordIndex
        actual = StructBusBurstInfo(lastWordIndex * wordIndexToAddrRatio, [])
        skippedWords = 0
        busBursts.append(actual)

        # for each part which should be download
        for info in structInfos:
            # check if space between fields is small enough
            startW = info.parts[0].wordIndex
            if startW - lastWordIndex - skippedWords > maxDummyWords:
                # space between useful fields is too big and we have to split burst into multiple smaller
                skippedWords += startW - lastWordIndex - 2
                actual = StructBusBurstInfo(startW * wordIndexToAddrRatio, [])
                busBursts.append(actual)

            if skippedWords != 0:
                # update word indexes after unused fields were potentially cut of
                for p in info.parts:
                    p.wordIndex -= skippedWords

            actual.fieldInfos.append(info)
            lastWordIndex = info.parts[-1].wordIndex

        return busBursts

    @staticmethod
    def sumOfWords(structBusBurstInfos):
        return structBusBurstInfos[-1].fieldInfos[-1].parts[-1].wordIndex + 1

    def __repr__(self):
        return "<StructBusBurstInfo addrOffset:%d>" % self.addrOffset



class BusFieldInfo(object):
    def __init__(self, access="rw", fieldInterface=None, disolveArray=False):
        """
        :param access: "r", "w" or "rw" describes access mode from bus side
        :param fieldInterface: interface for which this field was generated
        :param disolveArray: interpret this array interface as bunch of items
                             instead of single memory space
        """
        assert access in ['r', 'w', 'rw']
        self.access = access
        self.fieldInterface = fieldInterface
        self.disolveArray = disolveArray

def HStruct_selectFields(structT, fieldsToUse):
    """
    Select fields from structure (rest will become spacing)
    """
    template = []
    fieldsToUse = set(fieldsToUse)
    foundNames = set()

    for f in structT.fields:
        if f.name not in fieldsToUse:
            name = None
        template.append(HStructField(name, f.type))
        foundNames.add(name)

    assert fieldsToUse.issubset(foundNames)

    return HStruct(*template)




from hwt.hdlObjects.types.hdlType import HdlType
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class Array(HdlType):
    """
    vldMask and eventMask on Array_val instance is not used instead of that
    these flags on elements are used
    """
    def __init__(self, elmType, size):
        super(Array, self).__init__()
        self.elmType = elmType
        self.size = size

    def __hash__(self):
        return hash((self.elmType, self.size))

    def bit_length(self):
        try:
            itemSize = self.elmType.bit_length
        except AttributeError:
            itemSize = None
        if itemSize is None:
            raise TypeError("Can not determine size of array because item has not determinable size")
        
        s = self.size
        if isinstance(s, RtlSignalBase):
            s = s.staticEval()
        return s * itemSize()

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.arrayVal import ArrayVal
            cls._valCls = ArrayVal
            return cls._valCls



from hwt.synthesizer.exceptions import TypeConversionErr


class InvalidVHDLTypeExc(Exception):
    def __init__(self, vhdlType):
        self.vhdlType = vhdlType

    def __str__(self):
        variableName = self.variable.name
        return ("Invalid type, width is %s in the context of variable %s"
                % (str(self.vhdlType.getWidth()), variableName))

    def __repr__(self):
        return self.__str__()


class HdlType():
    def __init__(self):
        self.constrain = None

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash((self.name, self.constrain))

    def fromPy(self, v):
        return self.getValueCls().fromPy(v, self)

    def convert(self, sigOrVal, toType):
        if sigOrVal._dtype == toType:
            return sigOrVal

        try:
            c = self._convert
        except AttributeError:
            c = self.getConvertor()
            self._convertor = c

        return c(self, sigOrVal, toType)

    @classmethod
    def getConvertor(cls):
        return HdlType.defaultConvert

    def defaultConvert(self, sigOrVal, toType):
        raise TypeConversionErr("Conversion of %r of type \n%r to type %r is not implemented"
                                % (sigOrVal, self, toType))

    def __repr__(self):
        return "<HdlType %s>" % (self.__class__.__name__)



from hwt.hdlObjects.constants import READ, WRITE, NOP
from hwt.simulator.agentBase import SyncAgentBase
from hwt.simulator.shortcuts import oscilate


class BramPort_withoutClkAgent(SyncAgentBase):
    """
    :ivar requests: list of tuples (request type, address, [write data]) - used for driver
    :ivar data: list of data in memory, used for monitor
    :ivar mem: if agent is in monitor mode (= is slave) all reads and writes are performed on
        mem object
    """
    def __init__(self, intf, clk=None, rstn=None):
        super().__init__(intf, clk=clk, rstn=rstn, allowNoReset=True)

        self.requests = []
        self.readPending = False
        self.readed = []

        self.mem = {}

    def doReq(self, s, req):
        rw = req[0]
        addr = req[1]

        if rw == READ:
            rw = 0
            wdata = None
            self.readPending = True

        elif rw == WRITE:
            wdata = req[2]
            rw = 1
        else:
            raise NotImplementedError(rw)

        intf = self.intf
        s.w(rw, intf.we)
        s.w(addr, intf.addr)
        s.w(wdata, intf.din)

    def onReadReq(self, s, addr):
        """
        on readReqRecieved in monitor mode
        """
        self.requests.append((READ, addr))

    def onWriteReq(self, s, addr, data):
        """
        on writeReqRecieved in monitor mode
        """
        self.requests.append((WRITE, addr, data))

    def monitor(self, s):
        intf = self.intf

        yield s.updateComplete

        if self.requests:
            req = self.requests.pop(0)
            t = req[0]
            addr = req[1]
            assert addr._isFullVld(), s.now
            if t == READ:
                s.write(self.mem[addr.val], intf.dout)
            else:
                assert t == WRITE
                s.write(None, intf.dout)
                self.mem[addr.val] = req[2]

        if self.enable:
            en = s.read(intf.en)
            assert en.vldMask
            if en.val:
                we = s.read(intf.we)
                assert we.vldMask

                addr = s.read(intf.addr)
                if we.val:
                    data = s.read(intf.din)
                    self.onWriteReq(s, addr, data)
                else:
                    self.onReadReq(s, addr)

    def getDrivers(self):
        return [self.driver]

    def driver(self, s):
        intf = self.intf
        readPending = self.readPending
        if self.requests and self.enable:
            req = self.requests.pop(0)
            if req is NOP:
                s.w(0, intf.en)
                s.w(0, intf.we)
                self.readPending = False
            else:
                self.doReq(s, req)
                s.w(1, intf.en)
        else:
            s.w(0, intf.en)
            s.w(0, intf.we)
            self.readPending = False

        if readPending:
            yield s.updateComplete
            d = s.r(intf.dout)
            self.readed.append(d)


class BramPortAgent(BramPort_withoutClkAgent):
    def __init__(self, intf, clk=None, rstn=None):
        if clk is None:
            clk = intf.clk
        super().__init__(intf, clk=clk, rstn=rstn)

    def getDrivers(self):
        drivers = super(BramPortAgent, self).getDrivers()
        drivers.append(oscilate(self.intf.clk))
        return drivers



from hwt.simulator.agentBase import SyncAgentBase
from hwt.hdlObjects.constants import NOP


class HandshakedAgent(SyncAgentBase):
    """
    Simulation/verification agent for handshaked interface

    :attention: requires clk and rst/rstn signal
        (if you do not have any create simulation wrapper with it)
    """
    def __init__(self, intf, clk=None, rstn=None):
        super().__init__(intf, clk=clk, rstn=rstn)
        self.actualData = NOP
        self.data = []
        # these signals are extracted like this to make
        # agent more configurable
        self._rd = self.getRd()
        self._vld = self.getVld()

    def getRd(self):
        """get "ready" signal"""
        return self.intf.rd

    def getVld(self):
        """get "valid" signal"""
        return self.intf.vld

    def monitor(self, s):
        """
        Collect data from interface
        """
        if s.r(self.rst_n).val and self.enable:
            s.w(1, self._rd)

            yield s.updateComplete
            vld = s.r(self._vld)
            assert vld.vldMask, "valid signal for interface %r is in invalid state, this would cause desynchronization" % (self.intf)

            if vld.val:
                d = self.doRead(s)
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, read, %d: %r\n" % (
                                            self.intf._getFullName(), s.now, d))
                self.data.append(d)
        else:
            s.w(0, self._rd)

    def doRead(self, s):
        """extract data from interface"""
        return s.read(self.intf.data)

    def doWrite(self, s, data):
        """write data to interface"""
        s.w(data, self.intf.data)

    def driver(self, s):
        """Push data to interface

        set vld high and wait on rd in high then pass new data
        """
        if self.actualData is NOP and self.data:
            self.actualData = self.data.pop(0)

        do = self.actualData is not NOP

        if do:
            self.doWrite(s, self.actualData)
        else:
            self.doWrite(s, None)

        en = s.r(self.rst_n).val and self.enable
        if en and do:
            s.w(1, self._vld)
        else:
            s.w(0, self._vld)
            return

        yield s.updateComplete

        rd = s.r(self._rd)
        if en:
            assert rd.vldMask, "ready signal for interface %r is in invalid state, this would cause desynchronization" % (self.intf)
        if rd.val:
            if self._debugOutput is not None:
                self._debugOutput.write("%s, wrote, %d: %r\n" % (
                                           self.intf._getFullName(), s.now, self.actualData))
            if self.data:
                self.actualData = self.data.pop(0)
            else:
                self.actualData = NOP


class HandshakeSyncAgent(HandshakedAgent):
    """
    Simulation/verification agent for HandshakedSycn interface

    :attention: there is no data channel on this interface it is synchronization only
                and it actually does not have any meaningful data collected data in monitor
                mode are just values of simulation time when item was collected
    """

    def doWrite(self, s, data):
        pass

    def doRead(self, s):
        return s.now
